/*
 *   Copyright (c) - Ano de 2001               
 *      Anderson de Rezende Rocha - undersun@comp.ufla.br
 *      Joo Guilherme de Mendona Goulart - FALTA O MAIL
 *      Julio Cesar Alves - jcalves@comp.ufla.br
 *
 *      Disciplina de TECNICAS DE PROGRAMACAO
 *      Prof.: Bruno de Oliveira Schneider
 *
 *      Implementacao da Estrutura de dados GRAFO usando LISTA DE VIZINHANCAS
 *      Particularidades: Grafo Direcional com o intuito de demonstrar a atuacao do
 *      algoritmo de Caminho Minimo Num Grafo Direcional Aciclico
 *
*/

#ifndef CLGRAFO_H
#define CLGRAFO_H

#include "noh.h"            // Representa o NOH propriamente dito
#include "ligacao.h"     // Representa uma ligacao entre um NOH e outro
#include "lista.h"          // Usada para o algoritmo de Caminho Minimo

using namespace std;

// Classe Grafo propriamente dita
class clGrafo
{
    public:
        // Construtor e destrutor padrao
        clGrafo();
        ~clGrafo();

        // Operacoes sobre o GRAFO
        void InserirNoh(int id);
        bool InserirLigacao(int  idOrigem, int idDestino, int peso);
        void MostrarGrafoComoLista();

        // Algoritmo para busca do Caminho Minimo no Grafo
        bool CaMiGraDa(int idInicial);

    private:
        // Usada para destruir todos os nohs e ligacoes do grafo. Chamada pelo destrutor
        void DestruirGrafo();
        // Utilizada apenas para mostrar o grafo em forma de lista na tela
        void MostrarQualquerGrafoComoLista(clNoh* primeiroNohDoGrafo);
        // Devolve um noh com o primeiro dos ids, que se encontre a partir do nohInicial na
        // lista de nohs
        clNoh *EncontrarNoh(int id1, int id2, clNoh*nohInicial);
        // Noh cabeca da lista de nohs do grafo
        clNoh* primeiroNoh;

        // Funcoes utilizadas diretamente pelo Algoritmo CaMiGraDA

        // Funcao auxiliar para busca do caminho minimo
        void CaMiGraDaAux(clNoh* inicial);
        // Inicializacao das variaveis pai, estimativa...
        void IniciarCaMUO(clNoh* origem);
        // Faz o relaxamente, se possivel
        void RelaxarLigacao(clNoh* nohOrigem, clNoh* nohDestino, clLigacao* ligacao);
        // Utilizada para definir a hierarquia entre nohs
        void BuscaEmProfundidade(clLista<clNoh* > *listaTopologica);
        // Rotina Recursiva auxiliar 'a busca em profundidade
        void BuscaProfundidadeAux(clNoh* noh, clLista<clNoh* > *listaTopologica);
        // Ordenacao Topologica das ligacoes. Define a hierarquia entre os nohs
        // Devolve a lista hierarquica com ponteiros para os NOHS
        clLista<clNoh* > *OrdenarTopologicamente();

        // Apenas mostra o resultado da busca do caminho minimo
        void MostrarResultadosCaMiGraDa(clNoh* nohInicial);
        // Guarda o numero de acessos a memoria
        unsigned int nroAcessosMemoria;
};

//////////////////////////////////CONSTRUTOR DEFAULT/////////////////////////////////////
clGrafo::clGrafo()
{
    primeiroNoh = NULL;
}

//////////////////////////////////DESTRUTOR DEFAULT///////////////////////////////////////
clGrafo::~clGrafo()
{
    DestruirGrafo();
}

void clGrafo::DestruirGrafo()
{
    clNoh *nohAux;
    nohAux = primeiroNoh;
    clLigacao *ligacao;
    int arestas = 0;
    int nohs = 0;

    cout << "\n\nGRAFO sendo Destruido..." << endl;

    while( primeiroNoh != NULL )
    {
        nohAux = primeiroNoh;
        ligacao = primeiroNoh->PrimeiraLigacao();
        while( ligacao != NULL )
        {
            primeiroNoh->AlterarPrimeiraLigacao(ligacao->Proxima());
            arestas++;
            delete ligacao;
            ligacao = primeiroNoh->PrimeiraLigacao();
        }
        primeiroNoh = nohAux->Proximo();
        delete nohAux;
        nohs++;
    }
    cout << arestas << " LIGACOES destruidas" << endl;
    cout << nohs <<" NODO(s) destruidos\n" << endl;
}

///////////////////OPERACOES SOBRE O GRAFO (INSERCAO E REMOCAO)//////////////////////

// Insere os nohs ordenados por id, para que nao precise percorrer todos os nohs
// para saber se um id ja existe
void clGrafo::InserirNoh(int id)
{
    clNoh* novoNoh;
    novoNoh = new clNoh(id);

    if (primeiroNoh == NULL)
        primeiroNoh = novoNoh;
    else
    {
        clNoh* noh;
        noh = primeiroNoh;
        while (noh != NULL && id != noh->Id())
        {
            noh = noh->Proximo();
        }
        if (noh == NULL)
        {
            novoNoh->AlterarProximo(primeiroNoh);
            primeiroNoh = novoNoh;
        }
    }
}

bool clGrafo::InserirLigacao(int  idOrigem, int idDestino, int peso)
{
    clNoh* nohOrigem;
    clNoh* nohDestino;
    nohOrigem = EncontrarNoh(idOrigem, idDestino, primeiroNoh);
    if (nohOrigem != NULL)
    {
        if (nohOrigem->Id() != idOrigem)
        {
            nohDestino = nohOrigem;
            nohOrigem = EncontrarNoh(idOrigem, idDestino, nohDestino->Proximo());
        }
        else
            nohDestino = EncontrarNoh(idOrigem, idDestino, nohOrigem->Proximo());
        if (nohDestino != NULL && nohOrigem != NULL)
        {
            clLigacao* ligacaoOrigem;
            ligacaoOrigem = new
                        clLigacao(nohDestino, nohOrigem->PrimeiraLigacao(), peso);
            nohOrigem->AlterarPrimeiraLigacao(ligacaoOrigem);
            return true;
        }
        else
            return false; // nao foram encontrados idOrigem e idDestino
    }
    else
        return false;
}

//////////////////////////////////EXIBIR O GRAFO NA TELA/////////////////////////////////

void clGrafo::MostrarGrafoComoLista()
{
    MostrarQualquerGrafoComoLista(primeiroNoh);
}

/////////////////////CAMINHO MINIMO NUM GRAFO DIRECIONAL ACICLICO//////////////////////
bool clGrafo::CaMiGraDa(int idInicial)
{
    clNoh* noh;
    noh = EncontrarNoh(idInicial, idInicial, primeiroNoh);
    if (noh == NULL)
        return false;
    else
        CaMiGraDaAux(noh);
    return true;
}

void clGrafo::CaMiGraDaAux(clNoh* inicial)
{
    clNoh* noh;
    clLigacao* ligacao;
    clLista<clNoh* > *listaTopologica;

    nroAcessosMemoria = 0;
    listaTopologica = OrdenarTopologicamente();
    IniciarCaMUO(inicial);
    while (!listaTopologica->Vazia())
    {
        noh = listaTopologica->Cabeca();
        ligacao = noh->PrimeiraLigacao();
        while (ligacao != NULL)
        {
            // o vizinho eh ligacao->Destino()
            RelaxarLigacao(noh, ligacao->Destino(), ligacao);
            ligacao = ligacao->Proxima();
            nroAcessosMemoria += 9;
        }
        listaTopologica->RemoverCabeca();
        nroAcessosMemoria += 10;
    }
    nroAcessosMemoria += 2;
    MostrarResultadosCaMiGraDa(inicial);
    delete listaTopologica;
}

void clGrafo::IniciarCaMUO(clNoh* origem)
{
    clNoh* noh;
    noh = primeiroNoh;
    while (noh != NULL)
    {
        noh->AlterarEstimativa(INFINITO);
        noh->AlterarPai(NULL);
        noh = noh->Proximo();
    }
    origem->AlterarEstimativa(0);
    origem->AlterarPai(NULL);
}

void clGrafo::RelaxarLigacao(clNoh* nohOrigem, clNoh* nohDestino, clLigacao* ligacao)
{
    int soma;
    if (nohOrigem->Estimativa() == INFINITO)
    {
        soma = INFINITO;
        nroAcessosMemoria += 1;
    }
    else
    {
        soma = nohOrigem->Estimativa() + ligacao->Peso();
        nroAcessosMemoria += 5;
    }
    nroAcessosMemoria += 2;    // if
    if (nohDestino->Estimativa() > soma)
    {
        nohDestino->AlterarEstimativa(soma);
        nohDestino->AlterarPai(nohOrigem);
        nroAcessosMemoria += 6;
    }
    nroAcessosMemoria += 3;        // if
}

clLista<clNoh* > *clGrafo::OrdenarTopologicamente()
{
    clLista<clNoh* > *listaTopologica;
    listaTopologica = new clLista<clNoh* >;
    BuscaEmProfundidade(listaTopologica);
    nroAcessosMemoria += 2;
    return listaTopologica;
}

////////////////////////////////BUSCA EM PROFUNDIDADE////////////////////////////////
void clGrafo::BuscaEmProfundidade(clLista<clNoh* > *listaTopologica)
{
    clNoh* noh;
    noh = primeiroNoh;
    while (noh != NULL)
    {
        noh->AlterarCor(BRANCO);
        noh->AlterarPai(NULL);
        noh = noh->Proximo();
        nroAcessosMemoria += 8;
    }
    // tempo = 0;
    noh = primeiroNoh;
    while (noh != NULL)
    {
        if (noh->Cor() == BRANCO)
        {
            BuscaProfundidadeAux(noh, listaTopologica);
            nroAcessosMemoria += 3;
        }
        noh = noh->Proximo();
        nroAcessosMemoria += 6;
    }
    nroAcessosMemoria += 4;
}

void clGrafo::BuscaProfundidadeAux(clNoh* noh, clLista<clNoh* > *listaTopologica)
{
    clLigacao* ligacao;
    noh->AlterarCor(CINZA);
    ligacao = noh->PrimeiraLigacao();
    while (ligacao != NULL)
    {
        // o vizinho eh ligacao->Destino()
        if (ligacao->Destino()->Cor() == BRANCO)
        {
            ligacao->Destino()->AlterarPai(noh);
            BuscaProfundidadeAux(ligacao->Destino(), listaTopologica);
            nroAcessosMemoria += 8;
        }
        ligacao = ligacao->Proxima();
        // nao esta contando BRANCO
        nroAcessosMemoria += 7;   // do while tambem
    }
    noh->AlterarCor(PRETO);
    listaTopologica->InserirNaCabeca(noh);
    nroAcessosMemoria += 10;
}

///////////////////////// MOSTRAR O GRAFO COM LISTA NA TELA/////////////////////////
// Usando a saida padrao do sistema
void clGrafo::MostrarQualquerGrafoComoLista(clNoh* primeiroNohDoGrafo)
{
    cout << "\n MOSTRANDO O GRAFO COMO LISTA\n" << endl;
    clNoh* noh;
    clLigacao* ligacao;

    noh = primeiroNohDoGrafo;
    while (noh != NULL)
    {
        cout << noh->Id() << " -> ";
        if (noh->PrimeiraLigacao() != NULL)
        {
            ligacao = noh->PrimeiraLigacao();
            while (ligacao != NULL)
            {
                cout << ligacao->Destino()->Id() << "(" << ligacao->Peso() << ")" << " -> ";
                ligacao = ligacao->Proxima();
            }
        }
        cout << "NULL" << endl;
        noh = noh->Proximo();
    }
}

////////EXIBE NA SAIDA PADRAO O RESULTADO DA BUSCA DO CAMINHO MINIMO//////////
void clGrafo::MostrarResultadosCaMiGraDa(clNoh* nohInicial)
{
    clNoh* noh;
    clNoh* pai;
    pai = NULL;

    cout << "\t\nCaminhos Minimos a partir do noh " << nohInicial->Id() << ":\n" << endl;
    noh = primeiroNoh;
    while (noh != NULL)
    {
        cout << "Para o noh " << noh->Id() << ":" << endl;
        if (noh->Estimativa() != INFINITO)
        {
            cout << "Estimativa: " << noh->Estimativa() << " Caminho: ";
            cout << noh->Id();
            pai = noh->Pai();
            while (pai != NULL)
            {
                cout << " <- " << pai->Id();
                pai = pai->Pai();
            }
        }
        else
            cout << "Noh inalcancavel ";
        cout << endl;
        noh = noh->Proximo();
    }
    cout << "\nForam feitos " << nroAcessosMemoria;
    cout << " acessos a memoria para solucionar o problema" << endl;
}

///////////////////////////// ENCONTRAR UM NOH ESPECIFICO////////////////////////////
clNoh* clGrafo::EncontrarNoh(int id1, int id2, clNoh* nohInicial)
{
    // retorna NULL se o noh nao foi encontrado
    clNoh* noh;
    noh = nohInicial;
    while (noh != NULL && noh->Id() != id1 && noh->Id() != id2)
    {
        noh = noh->Proximo();
    }
    return noh;
}


#endif


