/*
 * clLista
 * Implementacao de uma lista usando templates
 * 
 * Obs.: A classe TValor deve definir:
 *       - Construtor de copia
 *       - operador =
 *       - operador <<
 *       - operador < (para usar o metodo InserirEmOrdem)
 *
 * (C) Bruno de Oliveira Schneider - DCC - UFLA, 7 de Junho de 2000
 */

#ifndef CLLISTA_H
#define CLLISTA_H

#include <iostream>

template <class TValor>
class clNohLista
// O operador = DEVE estar definido na classe TValor.
{
    public:
        clNohLista();
        ~clNohLista();
        clNohLista* Proximo();
        TValor Valor();
        void CopiarValor(const TValor &v);
        void Proximo(clNohLista* noh);
    protected:
        TValor* valor;
        clNohLista* proximo;
};

template <class TValor>
clNohLista<TValor>::clNohLista()
{
    valor = NULL;
    proximo = NULL;
}

template <class TValor>
clNohLista<TValor>::~clNohLista()
{
    if (valor != NULL)
        delete valor;
}

template <class TValor>
TValor clNohLista<TValor>::Valor()
{
    return *valor;
}

template <class TValor>
clNohLista<TValor>* clNohLista<TValor>::Proximo()
// retorna o valor do proximo
{
    return proximo;
}

template <class TValor>
void clNohLista<TValor>::Proximo(clNohLista* noh)
// muda o valor do proximo
{
    proximo = noh;
}

template <class TValor>
void clNohLista<TValor>::CopiarValor(const TValor &v)
// muda o valor do noh
{
    if (valor != NULL) 
        delete valor;
    valor = new TValor(v);
}

// Lista ====================================


template <class TValor>
class clLista
{
    friend ostream &operator<<<TValor>(ostream &output, const clLista<TValor> &l);
    public:
        clLista();
        clLista(const clLista<TValor> &lista);
        ~clLista();
        TValor Cabeca();
        clNohLista<TValor> *Primeiro() const;
        void InserirEmOrdem(const TValor &val);
        void InserirNaCabeca(const TValor &val);
        void InserirNoFim(const TValor &val);
        void RemoverCabeca();
        clNohLista<TValor> *Resto();
        void Limpar();
        bool Vazia();
        unsigned int Tamanho();
        void operator = (const clLista<TValor> &lista);
    protected:
        clNohLista<TValor> *primeiro;
};

template <class TValor>
clLista<TValor>::clLista()
{
    primeiro = NULL;
}

template <class TValor>
clLista<TValor>::clLista(const clLista<TValor> &lista)
{
    clNohLista<TValor> *noh;

    // criar a lista vazia...
    primeiro = NULL;    
    noh = lista.Primeiro();

    // ... e copiar cada elemento
    while (noh != NULL)
    {
        InserirNoFim(noh->Valor());
        noh = noh->Proximo();
    }
}

template <class TValor>
clLista<TValor>::~clLista()
{
    Limpar(); //remover todos os nohs
}

template <class TValor>
bool clLista<TValor>::Vazia()
{
    return (primeiro == NULL);
}

template <class TValor>
clNohLista<TValor> *clLista<TValor>::Primeiro() const
{
    return primeiro;
}

template <class TValor>
TValor clLista<TValor>::Cabeca()
{
    return primeiro->Valor();
}

template <class TValor>
clNohLista<TValor> *clLista<TValor>::Resto()
{
    return primeiro->Proximo();
}

template <class TValor>
void clLista<TValor>::Limpar()
// Apaga todos elementos da lista, tornando-a vazia
{
    clNohLista<TValor> *aux;

    // destruir todos os nohs da lista
    aux = primeiro;
    while(aux != NULL)
    {
        primeiro = aux->Proximo();
        delete aux;
        aux = primeiro;
    }
}

template <class TValor>
void clLista<TValor>::InserirNoFim(const TValor &val)
// Insere UMA COPIA do valor passado no fim da lista
{
    clNohLista<TValor> *atual;
    clNohLista<TValor> *anterior;
    
    atual = primeiro;
    anterior = NULL;
    
    while (atual != NULL)
    {
        anterior = atual;
        atual = atual->Proximo();
    }
    atual = new clNohLista<TValor>;
    atual->CopiarValor(val);
    if (anterior == NULL)
        primeiro = atual;
    else
        anterior->Proximo(atual);
}

template <class TValor>
void clLista<TValor>::InserirNaCabeca(const TValor &val)
// Insere UMA COPIA do valor passado na cabeca da lista
{
    clNohLista<TValor> *noh;

    noh = new clNohLista<TValor>;
    noh->CopiarValor(val);
    noh->Proximo(primeiro); // apontar para o primeiro
    primeiro = noh; // mudar o primeiro
}

template <class TValor>
void clLista<TValor>::RemoverCabeca()
// Remove a cabeca da lista
{
    clNohLista<TValor> *noh;

    noh = primeiro;
    primeiro = primeiro->Proximo();
    delete noh;
}

template <class TValor>
void clLista<TValor>::operator = (const clLista<TValor> &lista)
// copia uma lista sobre outra
{
    clNohLista<TValor> *noh;
    
    Limpar(); // Apaga tudo que tinha na lista antes
    noh = lista.Primeiro();
    while (noh != NULL)
    {
        InserirNoFim(noh->Valor());
        noh = noh->Proximo();
    }
}

template <class TValor>
void clLista<TValor>::InserirEmOrdem(const TValor &val)
// Este metodo supoe que a lista esteja previamente ordenada
{
    clNohLista<TValor> *noh;
    clNohLista<TValor> *novoNoh;
    clNohLista<TValor> *anterior;

    noh = primeiro;
    anterior = NULL;
    while ((noh != NULL) && (noh->Valor() < val))
    {
        anterior = noh;
        noh = noh->Proximo();
    }
    novoNoh = new clNohLista<TValor>;
    novoNoh->CopiarValor(val);
    novoNoh->Proximo(noh);
    if (anterior == NULL)
        primeiro = novoNoh;
    else
        anterior->Proximo(novoNoh);
}

template <class TValor>
unsigned int clLista<TValor>::Tamanho()
// Retorna a quantidade de elementos na lista
{
    unsigned int tamanho=0;
    clNohLista<TValor> *noh;    
    
    noh = primeiro;
    while (noh != NULL)
    {
        tamanho++;
        noh = noh->Proximo();
    }
    return tamanho;
}

template <class TValor>
ostream &operator<<(ostream &output, const clLista<TValor> &l)
// operador de padrao de saida, imprime a lista com parenteses e ;
{
    clNohLista<TValor> *noh;

    output << "(";
    noh = l.primeiro;
    if (noh != NULL)
    {
        output << noh->Valor();
        noh = noh->Proximo();
    }
    while (noh != NULL)
    {
        output << "; " << noh->Valor();
        noh = noh->Proximo();
    }
    output << ")";
    return output;
}

#endif
